#include "fpscommon.h"
char newarg[120][MAXPATHLEN - 64];
int newarg_index = 0;

/*****************************************************/

reset_process_line(reset)
struct ps_line *reset;
{
   reset->user[0]    = (char)NULL; 
   reset->pid[0]     = (char)NULL;  
   reset->cpu[0]     = (char)NULL;  
   reset->mem[0]     = (char)NULL;
   reset->sz[0]      = (char)NULL;   
   reset->rss[0]     = (char)NULL;  
   reset->tt[0]      = (char)NULL;   
   reset->stat[0]    = (char)NULL;
   reset->eng[0]     = (char)NULL;  
   reset->time[0]    = (char)NULL; 
   reset->command[0] = (char)NULL;
}

/*************************************************************/
/* Macros to aid in processing a line from ps. */
#define ReadToken(inbuf,maxlen,val) \
{ int __i; for ( __i=0; __i<maxlen && !isspace(*inbuf); __i++) \
    val[__i] = *inbuf++; val[__i] = 0; }
#define SkipToken(inbuf,maxlen) \
{ int __i; for ( __i=0; __i<maxlen && !isspace(*inbuf); __i++) inbuf++;}
#define SkipSpace( inbuf ) \
    while (isspace(*inbuf)) inbuf++;
#define ReadTokenAndSkip( inbuf, maxlen, val ) \
    ReadToken( inbuf, maxlen, val ); SkipSpace( inbuf );

/* Length of the fields that are sometimes different */
#ifdef SUT_rs6000
#define PS_TTY 6
#else
#define PS_TTY 2
#endif

#ifdef PS_SYSV
get_process_line(in_buf, u)
char *in_buf;
struct ps_line *u;
{  
int i;
/* Skip flags */
in_buf += 2;
SkipSpace( in_buf );

ReadTokenAndSkip( in_buf, 3, u->stat );
ReadTokenAndSkip( in_buf, 8, u->user );
ReadTokenAndSkip( in_buf, 6, u->pid );
SkipToken( in_buf, 6 );   /* ppid */
SkipSpace( in_buf );
SkipToken( in_buf, 3 );   /* utilization */
SkipSpace( in_buf );
SkipToken( in_buf, 3 );   /* priority */
SkipSpace( in_buf );
SkipToken( in_buf, 2 );   /* Nice */
SkipSpace( in_buf );
SkipToken( in_buf, 2 );   /* Processor running process */
SkipSpace( in_buf );
for (i=0; *in_buf && *in_buf != ':'; i++) u->sz[i] = *in_buf++;
u->sz[i] = 0; if (*in_buf) in_buf++;
ReadToken( in_buf, 6, u->rss );

/* The next field is either blank or has a value in it (!) */
/* We just skip over 14 - strlen(u->rss) */
in_buf += 14 - strlen(u->rss);

SkipSpace( in_buf );
ReadTokenAndSkip( in_buf, 8, u->eng );
ReadTokenAndSkip( in_buf, 6, u->tt );
ReadTokenAndSkip( in_buf, 8, u->time );

sscanf(in_buf, "%[^\n]", u->command);     /* COMMAND */
in_buf += strlen(u->command);
}
#else
get_process_line(in_buf, u)
char *in_buf;
struct ps_line *u;
{  

   ReadTokenAndSkip( in_buf, 8, u->user );

   ReadTokenAndSkip( in_buf, 6, u->pid );

   if(*in_buf != 'Z')
   {
      ReadTokenAndSkip( in_buf, 5, u->cpu );

      /* Actual sizes may exceed the available area */
      ReadTokenAndSkip( in_buf, 4, u->mem );
      
      ReadTokenAndSkip( in_buf, 4, u->sz );

      ReadToken( in_buf, 4, u->rss );
      /* Skip any remaining digits; this isn't correct, but there
	 isn't any other way to get data (should actually change
	 sz and rss to '*') */
      while(isdigit(*in_buf)) in_buf++;
      SkipSpace( in_buf );

      ReadTokenAndSkip( in_buf, PS_TTY, u->tt );
   }

#ifdef SUT_rs6000
   sscanf(in_buf, "%2[A-Z ]", u->stat);
   in_buf +=  strlen(u->stat);
   if(strcmp(u->stat, "Z "))
   {
      SkipSpace( in_buf );
      if(isalpha(*in_buf))
	 sscanf(in_buf, "%6[A-z 0-9]", u->eng);
      else
	 sscanf(in_buf, "%8[A-z :0-9]", u->eng);
      in_buf += strlen(u->eng);
   }
#else
   sscanf(in_buf, "%4[A-Z ]", u->stat);     /* STAT */
   in_buf += strlen(u->stat);
   if(strcmp(u->stat, "Z "))
   {
      SkipSpace( in_buf );

      if(isalpha(*in_buf))
	 sscanf(in_buf, "%6[A-z 0-9]", u->eng);
      else                                           /* START */
	 sscanf(in_buf, "%5[A-z :0-9]", u->eng);
      in_buf += strlen(u->eng);
   }
#endif
   SkipSpace( in_buf );

   ReadTokenAndSkip( in_buf, 7, u->time );

   sscanf(in_buf, "%[^\n]", u->command);     /* COMMAND */
   in_buf += strlen(u->command);

}
#endif

/*************************************************************/

print_process_line(u)
struct ps_line *u;
{
#if defined(SUT_rs6000)
   fprintf( stdout, "%-8s %6s %5s %4s %4s %4s %6s %2s %8s %7s %s\n",
	  u->user, u->pid, u->cpu, u->mem, u->sz, u->rss,
	  u->tt, u->stat, u->eng, u->time, u->command);
#elif defined(PS_SYSV)
   fprintf( stdout, "%-8s %6s %4s %4s %6s %2s %8s %7s %s\n",
	  u->user, u->pid, u->sz, u->rss,
	  u->tt, u->stat, u->eng, u->time, u->command);
#else   
   fprintf( stdout, "%-8s %6s %5s %4s %4s %4s %2s %4s %6s %6s %s\n",
	  u->user, u->pid, u->cpu, u->mem, u->sz, u->rss,
	  u->tt, u->stat, u->eng, u->time, u->command);
#endif
}
/*************************************************************/

#include <signal.h>
#ifdef HAVE_SETPRIORITY
#include <sys/resource.h>
#endif

static int NiceValue = -1;
static int NiceIncrement = 1;

int GetNiceValue( args )
char ***args;
{
char **Argv = *args;
char *p;

Argv++;
p = Argv[0];
if (!p) return;
if (*p == '+') {
    p++;
    if (isdigit(p)) NiceIncrement = atoi(p);
    }
else if (*p == '-') {
    p++;
    if (isdigit(p)) NiceIncrement = - atoi(p);
    }
else if (isdigit(p)) {
    NiceIncrement = 0;
    NiceValue     = atoi(p);
    }
}

/* Signal value to use */
static int sigvalue = SIGINT;
int GetKillSignal( args )
char ***args;
{
char **Argv = *args;
char *p;
char *argn;

Argv++;
if (!Argv[0]) return;
    
p = Argv[0];
if (p && *p == '-') p++;
if (isdigit(*p)) {
    sigvalue = atoi(p);
    if (sigvalue <= 0) sigvalue = SIGINT;
    }
else if (isalnum(*p)) {
    while (isalnum(*p)) p++;
    if (*p) return;
    if (strncmp(Argv[0], "SIG", 3 ) == 0) argn = Argv[0] + 3;
    else                                  argn = Argv[0];
    if (strcmp( argn, "HUP") == 0)
	sigvalue = SIGHUP;
    else if (strcmp( argn, "INT") == 0)
	sigvalue = SIGINT;
    else if (strcmp( argn, "QUIT") == 0)
	sigvalue = SIGQUIT;
    else if (strcmp( argn, "ILL" ) == 0)
	sigvalue = SIGILL;
    else if (strcmp( argn, "TRAP") == 0)
	sigvalue = SIGTRAP;
    else if (strcmp( argn, "IOT") == 0)
	sigvalue = SIGIOT;
    else if (strcmp( argn, "ABRT") == 0)
	sigvalue = SIGABRT;
    else if (strcmp( argn, "EMT") == 0)
	sigvalue = SIGEMT;
    else if (strcmp( argn, "FPE") == 0)
	sigvalue = SIGFPE;
    else if (strcmp( argn, "KILL") == 0)
	sigvalue = SIGKILL;
    else if (strcmp( argn, "BUS") == 0)
	sigvalue = SIGBUS;
    else if (strcmp( argn, "SEGV") == 0)
	sigvalue = SIGSEGV;
    else if (strcmp( argn, "SYS") == 0)
	sigvalue = SIGSYS;
    else if (strcmp( argn, "PIPE") == 0)
	sigvalue = SIGPIPE;
    else if (strcmp( argn, "ALRM") == 0)
	sigvalue = SIGALRM;
    else if (strcmp( argn, "TERM") == 0)
	sigvalue = SIGTERM;
    else if (strcmp( argn, "URG") == 0)
	sigvalue = SIGURG;
    else if (strcmp( argn, "STOP") == 0)
	sigvalue = SIGSTOP;
    else if (strcmp( argn, "TSTP") == 0)
	sigvalue = SIGTSTP;
    else if (strcmp( argn, "CONT") == 0)
	sigvalue = SIGCONT;
    else if (strcmp( argn, "CHLD") == 0)
	sigvalue = SIGCHLD;
    else if (strcmp( argn, "TTIN") == 0)
	sigvalue = SIGTTIN;
    else if (strcmp( argn, "TTOU") == 0)
	sigvalue = SIGTTOU;
    else if (strcmp( argn, "IO") == 0)
	sigvalue = SIGIO;
    else if (strcmp( argn, "XCPU") == 0)
	sigvalue = SIGXCPU;
    else if (strcmp( argn, "XFSZ") == 0)
	sigvalue = SIGXFSZ;
    else if (strcmp( argn, "VTALRM") == 0) 
	sigvalue = SIGVTALRM;
    else if (strcmp( argn, "PROF") == 0)
	sigvalue = SIGPROF;
    else if (strcmp( argn, "WINCH") == 0) 
	sigvalue = SIGWINCH;
#ifdef SIGLOST
    else if (strcmp( argn, "LOST") == 0)
	sigvalue = SIGLOST;
#endif
    else if (strcmp( argn, "USR1") == 0)
	sigvalue = SIGUSR1;
    else if (strcmp( argn, "USR2") == 0)
	sigvalue = SIGUSR2;
    else 
	fprintf( stderr, "Unrecognized signal value %s\n", Argv[0] );
    }
}

print_pids(u)
struct ps_line *u;
{
   int pid, err;
   
   if(KILL) {
       pid = atoi( u->pid );
       if (pid > 0) err = kill( pid, sigvalue );
       if (err == 0) {
	   fprintf( stdout, "process %s killed...\n", u->pid);
	   }
       }
#ifdef HAVE_SETPRIORITY
   else if (NICE) {
       pid = atoi( u->pid );
       if (pid > 0) 
	   if (NiceValue >= 0)
	       err = setprioity( PRIO_PROCESS, pid, NiceValue );
	   else 
	       err = setprioity( PRIO_PROCESS, pid, 
			  getpriority( PRIO_PROCESS, pid ) + NiceIncrement );
       }
#endif
   else if (EXEC) {
       RunProgram( u );
       }
   else {
      fprintf( stdout, "%s\n", u->pid );
      }
}
/*************************************************************/

/* Do exec of the pid */
static char execline[1024];
int RunProgram( u )
struct ps_line *u;
{
char pgmline[1024];
sprintf( pgmline, execline, u->pid );
/* fprintf( stderr, "%s\n", pgmline ); */
system( pgmline );
}

/* Get the execline.  Replace {} with %s (for the pid) */
int GetExecline( args )
char ***args;
{
char **Argv = *args;
execline[0] = 0;
Argv++;
while (Argv[0]) {
    if (Argv[0][0] == ';') break;
    if (strcmp( Argv[0], "{}" ) == 0) 
	strcat( execline, "%s" );
    else
	strcat( execline, Argv[0] );
    strcat( execline, " " );
    Argv[0] = 0;
    Argv++;
    }
}
/*************************************************************/

int UseTail = 0;
int set_use_tail()
{
UseTail = 1;
}

compare_name(in_command, u)                /* match the command name */
char *in_command;
struct ps_line *u;
{
   char *p1, *p2;

   if (UseTail) return compare_tail( in_command, u );
   /* Note that in_command may have a trailing blank, and u->command
      may contain multiple tokens.  Match the first token only */
   p1 = in_command;
   p2 = u->command;
   while (*p1 && *p2 && *p1 == *p2 && *p1 != ' ' && *p2 != ' ') {
       p1++;
       p2++;
       }
   /* Found a difference.  Why? */
   if (*p2 == ' ' && *p1 == ' ') return 1;
   if (*p2 == ' ' && !*p1) return 1;
   if (*p1 == ' ' && !*p2) return 1;
   if (*p1 != *p2) return 0;
   return(strcmp(in_command, u->command) ? 0 : 1);
}

/*************************************************************/

compare_tail(in_tail, u)                   /* match the tail of the command */
char *in_tail;
struct ps_line *u;
{
   char *p, *plast, *p1, *p2;
   /* Find last '/' in u->command */
   p = plast = u->command;
   while (p && *p && *p != ' ') {
       plast = p;
       while (*p && *p != '/' && *p != ' ') p++;
       if (*p == '/') 
	   plast = ++p;
       }
   
   p1 = in_tail;
   p2 = plast;
   while (*p1 && *p2 && *p1 == *p2 && *p1 != ' ' && *p2 != ' ') {
       p1++;
       p2++;
       }
   /* Found a difference.  Why? */
   if (*p2 == ' ' && *p1 == ' ') return 1;
   if (*p2 == ' ' && !*p1) return 1;
   if (*p1 == ' ' && !*p2) return 1;
   if (*p1 != *p2) return 0;
   return(strcmp(in_tail, plast) ? 0 : 1);
}

/*************************************************************/

compare_owner(in_owner, u)                 /* match the owner of the process */
char *in_owner;
struct ps_line *u;
{
   return(strcmp(in_owner, u->user) ? 0 : 1);
}

/*************************************************************/

compare_rtime(in_run_time, u)             /* match the running time */
char *in_run_time;
struct ps_line *u;
{
   int value;
   int in_sec, u_sec;
   int m, s;
   sscanf( in_run_time, "%d:%d", &m, &s );
   in_sec = s + 60 * m;
   sscanf( u->time, "%d:%d", &m, &s );
   u_sec  = s + 60 * m;
   return in_sec < u_sec;
}

/*************************************************************/
#include "time.h"
/* Not correct yet */
int get_mins_from_string( str )
char *str;
{
char *p, *p1, buf[30];
int  t[3], tcnt;

strcpy(buf,str);
/* If the string has the form %d:%d:%d, it is days, hours, mins, 
   else if it is %d:%d, it is hours, mins, else if it is %d, it
   is mins */
p    = buf;
tcnt = 0;
while (*p && tcnt < 3) {
    p1 = p;
    while (*p1 && *p1 != ':') p1++;
    *p1 = 0;
    t[tcnt++]  = atoi(p);
    p   = p1 + 1;
    }
switch (tcnt) {
    case 0: return 0;
    case 1: return t[0];
    case 2: return t[1] + 60 * t[0];
    case 3: return t[2] + 60 * (t[1] + 24 * t[2]);
    default: return 0;
    }
}
compare_stime(in_start_time,u)            /* match the start time */
char *in_start_time;
struct ps_line *u;
{
struct tm t_in, t_u, *t_local;
time_t cur_time = time( (time_t*)0 );
int    days, hours, mins, runmins;

/* Get the time limits */
mins = get_mins_from_string( in_start_time );

/* Get the local time */
t_local = localtime( &cur_time );

/* Convert to the latest acceptable starting time */
/* not finished ... */

/* Get the date or time that it started */
if (isdigit( u->eng[0] )) {
#if HAS_STRPTIME
    strptime( u->eng, "%H:%M", &t_u );
    t_u.tm_hour = t_u.tm_min = 0;
#else    
    t_u = *t_local;
    sscanf( u->eng, "%d:%d", &t_u.tm_hour, &t_u.tm_min );
#endif
    }
else {
    /* %h = month (name), %d = day of month */
#ifdef HAS_STRPTIME
    strptime( u->eng, "%h %d", &t_u );
    t_u.tm_yday = t_local->tm_yday;
#else    
    t_u = *t_local;
#endif
    }


/* local time - start time == run time */
days = t_local->tm_yday - t_u.tm_yday;
hours = (t_local->tm_hour - t_u.tm_hour) + 24 * days;
mins  = (t_local->tm_min - t_u.tm_min) + 60 * hours;
runmins = mins;

mins = t_in.tm_min + 60 * (t_in.tm_hour + 24 * t_in.tm_mday);
return (runmins < mins);
}


/*************************************************************/

compare_tty(in_tty, u)                    /* match the controlling tty */
char *in_tty;
struct ps_line *u;
{
   return(strcmp(in_tty, u->tt) ? 0 : 1);
}

/*************************************************************/

compare_state(in_state, u)                /* match the state */
char *in_state;
struct ps_line *u;
{
   return(strpbrk(u->stat, in_state) ? 1 : 0);
}

/*************************************************************/

fill_entry_list()
{
     int al_index = 0;
     int el_index = 0;
      
     while(strlen(newarg[al_index]))
     {
	if((newarg[al_index][0] == '-' )||(newarg[al_index][0] == '!'))
	{
	   if(!strcmp(newarg[al_index], "-or"))
	      entry_list[el_index].command_type = OR;
	   else if(!strcmp(newarg[al_index], "-and"))
	      entry_list[el_index].command_type = AND;
	   else if(!strcmp(newarg[al_index], "!"))
	      entry_list[el_index].command_type = NOT;
	   else 
	      entry_list[el_index].command_type = SWITCH;
	   if(!strcmp(newarg[al_index], "-tn")) {
	       UseTail = 1; 
	       }
	}
	if(entry_list[el_index].command_type == SWITCH)
	{
	   if(!strcmp(newarg[al_index], "-n"))
	   {
	      entry_list[el_index].fpointer = (compare_name);
	      strcat(newarg[al_index + 1], " ");
	   }	   
	   else if(!strcmp(newarg[al_index], "-o"))
	      entry_list[el_index].fpointer = (compare_owner);
	   else if(!strcmp(newarg[al_index], "-pty"))
	      entry_list[el_index].fpointer = (compare_tty);
	   else if(!strcmp(newarg[al_index], "-rtime"))
	      entry_list[el_index].fpointer = (compare_rtime);
	   else if(!strcmp(newarg[al_index], "-stime"))
	      entry_list[el_index].fpointer = (compare_stime);
	   else if(!strcmp(newarg[al_index], "-r"))
	      entry_list[el_index].fpointer = (compare_state);
	   al_index++;
	   entry_list[el_index].command_type = STRING;
	}
	strcpy(entry_list[el_index].arg, newarg[al_index]);
	al_index++;
	el_index++;
     }
}

/*************************************************************/

int
evaluate_process_line(u)
struct ps_line *u;
{
   int el_index = 0;
   int a = 0;
   int b = 0;
   
   S_INDEX = 0;
   while(entry_list[el_index].command_type != NONE)
   {
      if(entry_list[el_index].command_type == STRING)
      {
         if (!entry_list[el_index].fpointer) {
	     fprintf( stderr, "Unrecognized arguments!\n" );
	     exit(1);
	     }
	 STACK[S_INDEX] = entry_list[el_index].fpointer(entry_list[el_index].arg, u);
	 S_INDEX++;
      }
      else if(entry_list[el_index].command_type == NOT)
      {
	 if(STACK[S_INDEX - 1] == 1)
	    STACK[S_INDEX - 1] = 0;
	 else
	    STACK[S_INDEX - 1] = 1;
      }
      else if(entry_list[el_index].command_type == AND)
      {
	 S_INDEX--;
	 a = STACK[S_INDEX];
	 S_INDEX--;
	 b = STACK[S_INDEX];
	 if(a && b)
	    STACK[S_INDEX] = 1;
	 else
	    STACK[S_INDEX] = 0;
	 S_INDEX++;
      }
      else if(entry_list[el_index].command_type == OR)
      {
	 S_INDEX--;
	 a = STACK[S_INDEX];
	 S_INDEX--;
	 b = STACK[S_INDEX];
	 if(a||b)
	    STACK[S_INDEX] = 1;
	 else
	    STACK[S_INDEX] = 0;
	 S_INDEX++;
      }
      el_index++;
   }
   return(STACK[0]);
}

/*************************************************************/

int
get_token(arg)
char ***arg;
{
   TOKEN_CODE temptoken;
   
   if(***arg == '-')
   {
      if(!strcmp(**arg, "-or"))
	 temptoken = OR;
      else if(!strcmp(**arg, "-and"))
	 temptoken = AND;
      else if(!strcmp(**arg, "-print"))
	 temptoken = OPTION;
      else if(!strcmp(**arg, "-sort"))
	 temptoken = OPTION;
      else if(!strcmp(**arg, "-exec")) {
	 temptoken = OPTION;
	 /* Eat args into an evaluation place */
	 GetExecline( arg );
	 }
      else if(!strcmp(**arg, "-kill")) {
	 temptoken = OPTION;
	 /* Eat next arg as the kill signal value */
	 GetKillSignal( arg );
	 }
      else if(!strcmp(**arg, "-tn"))
	 temptoken = OPTION;
      else if(!strcmp(**arg, "-nice")) {
	 temptoken = OPTION;
	 /* Eat next arg as the nice increment or absolute value */
	 GetNiceValue( arg );
	 }
      else
	 temptoken = SWITCH;
   }
   else if(***arg == '(')
      temptoken = LPAREN;
   else if(***arg == ')')
      temptoken = RPAREN;
   else if(***arg == '!')
      temptoken = NOT;
   else 
      temptoken = STRING;
   evaluate_token(temptoken, arg);
   return(temptoken);
}

/*************************************************************/

evaluate_token(tok, etarg)
TOKEN_CODE tok;
char ***etarg;
{
   TOKEN_CODE temp;
   
   switch(tok)
   {
   case AND:
      (*etarg)++;
      get_token(etarg);
      strcpy(newarg[newarg_index], "-and");
      newarg_index++;
      break;
   case OR:
      (*etarg)++;
      get_token(etarg);
      strcpy(newarg[newarg_index], "-or");
      newarg_index++;
      break;
   case NOT:
      (*etarg)++;
      get_token(etarg);
      strcpy(newarg[newarg_index], "!");
      newarg_index++;
      break;
   case LPAREN:
      (*etarg)++;
      do
      {
	 temp = get_token(etarg);    /* recursively match left, ... */
      }                              /* ...right parentheses        */
      while(temp != RPAREN);
      break;
   case RPAREN:
      (*etarg)++;
      break;
   case STRING:
      strcpy(newarg[newarg_index], **etarg);
      (*etarg)++;
      newarg_index++;
      break;
   case SWITCH:
      strcpy(newarg[newarg_index], **etarg);
      newarg_index++;
      (*etarg)++;
      get_token(etarg);
      break;
   case OPTION:
      (*etarg)++;
      break;
   }
}
